import numpy as np
import random as rn
import pandas as pd
import itertools
import math

def distr(x):
    cumsum = []
    for i in np.arange(x.shape[0]):
        cumsum.append(sum(x[:i+1]))
    k = 1
    ran = rn.random()
    while ran > cumsum[k-1]:
        k = k + 1
    return k

def new_agentset(N,cards):
    print('Creating agentset of', N, 'agents...')
    agentset = []
    for i in range(N):
        agentset.append(agent(i,cards))
    return agentset
        
class new_set:
    def __init__(self, N):
        print('Creating new set of', N, 'cards...')
        self.N = N
        self.links = np.arange(N**2)
        rn.shuffle(self.links)
        self.links = self.links.reshape(N,N)
        for i in np.arange(N):
            for j in np.arange(N):
                if self.links[i,j] > (3*N):
                    self.links[i,j] = 0
                elif self.links[i,j] < (2.5*N):
                    self.links[i,j] = 1
                else: self.links[i,j] = 2
        for i in np.arange(N):
            self.links[i,i] = 1
            for j in np.arange(N):
                self.links[i,j] = self.links[j,i]
        x = np.linspace(1,30,30)
        y = np.array([math.exp(-x[i]/10) for i in range(len(x))])
        y = y/sum(y)
        self.features = np.zeros((N,1))
        for i in range(N):
            self.features[i,0] = distr(y)
        self.complete_set = pd.DataFrame(data = self.features, columns=['power'])
        self.prices = []
        for i in range(N):
            self.prices.append([])
            
    def booster(self):
        booster_pack = []
        for i in range(10):
            booster_pack.append(rn.randint(0,self.N-1))
        return booster_pack

class agent:
    def __init__(self,number,cards):
        self.number = number
        self.decklist = []
        self.k = (-18+5*math.sqrt(13))/27
        x = np.zeros((len(cards.complete_set),5))
        self.collection = pd.DataFrame(data = x, columns=['Q','Quality','MaxPrice','MinPrice','ite'])
        self.set_price()
        self.money = 30

    def buy_booster(self,cards):
        if self.money > 5:
            self.money -= 5
            booster = cards.booster()
            for i in np.arange(len(booster)):
                self.collection['Q'][booster[i]] += 1

    def set_price(self):
        for i in np.arange(self.collection.shape[0]):
            self.collection['MinPrice'][i] = rn.randint(1,20)
            self.collection['MaxPrice'][i] = self.collection['MinPrice'][i]*9/11
            self.collection['MaxPrice'][i] = round(self.collection['MaxPrice'][i],2)
            self.collection['Quality'][i] = 1

    def select_card(self):
        x = rn.random()
        x = x*sum(self.collection.Quality)
        k = 1
        while sum(self.collection.Quality[0:k]) <= x:
            k = k + 1
        return k - 1
        
    def buy_card(self,agentset,cards,delta,verbose = False,buyer_trials = 3):
        number_of_trials = 0
        buyer_desired_stock = 3
        selected_card = self.select_card()
        while self.collection.Q[selected_card] > 2:
            selected_card = self.select_card()
        trade = False
        buyer_desired_stock = 3
        while number_of_trials < buyer_trials and trade == False:
            seller = rn.choice(agentset)
            diff = self.collection.MaxPrice[selected_card] - seller.collection.MinPrice[selected_card]
            Q = seller.collection.Q[selected_card]
            if diff > 0 and Q > 0:
                price = (self.collection.MaxPrice[selected_card]+seller.collection.MinPrice[selected_card])/2
                #price = self.collection.MaxPrice[selected_card] - rn.uniform(0,diff)
                if self.money > price:
                    self.money = self.money - price
                    seller.money = seller.money + price
                    self.collection.Q[selected_card] += 1
                    seller.collection.Q[selected_card] -= 1
                    self.collection.ite[selected_card] = 0
                    seller.collection.ite[selected_card] = 0
                    trade = True
                    cards.prices[selected_card].append(price)
                    if verbose:
                        print('Agent', self.number, 'trade card number', selected_card, 'for the price of', price, 'with agent', seller.number)
            number_of_trials += 1
            if self.collection.Q[selected_card] < buyer_desired_stock:
                self.collection.MaxPrice[selected_card] += delta
                self.collection.MinPrice[selected_card] = self.collection.MaxPrice[selected_card] * 11/9

    def make_deck(self):
        Q = list(self.collection.Q)
        for i in range(len(Q)):
            if Q[i] > 3:
                Q[i] = 3
        self.decklist = []
        while len(self.decklist) < 30:
            x = self.select_card()
            if Q[x] > 0:
                self.decklist.append(x)
                Q[x] = Q[x] - 1

    def update_Quality(self,top_cards):
        for i in top_cards:
            ran = rn.random()
            if ran < 0.75:
                self.collection.Quality[i] += 1

    def update_MinPrice(self,delta,tt):
        M = max(self.collection.Quality)
        a = -1/(self.k-1)
        b = -M/math.log(self.k)
        c = 1-a
        for i in range(len(self.collection)):
            t = self.collection.ite[i]
            p = 1/(1+math.exp(-t+tt))
            ran = rn.random()
            if ran < p:
                x = self.collection.Quality[i]
                y = a*math.exp(-x/b)+c
                ran2 = rn.random()
                if ran2 < y:
                    self.collection.MaxPrice[i] -= delta
                    self.collection.MinPrice[i] = self.collection.MaxPrice[i] * 11/9
                    self.collection.ite[i] = 0
                    if self.collection.MaxPrice[i] < 0:
                        self.collection.MaxPrice[i] = 0.05
                        self.collection.MinPrice[i] = self.collection.MaxPrice[i] * 11/9

    def update_ite(self):
        for i in range(len(self.collection)):
            self.collection.ite[i] += 1

    def update_money(self):
        self.money += 10

    def take_decision_of_buying_booster(self,cards):
        c = len(self.collection[self.collection.Q > 0])
        p = 1/(1+math.exp(-(c-60)/30))
        ran = rn.random()
        if ran > p:
            self.buy_booster(cards)
                    
class game:
    def __init__(self,cards):
        self.cards = cards

    def match(self,player1,player2):
        points1 = 0
        points2 = 0
        for i in range(3):
            result = self.submatch(player1.decklist,player2.decklist)
            if result == 0: points1 = points1 + 1
            if result == 1: points2 = points2 + 1
        if points1 > points2: return 1
        elif points1 < points2: return 2
        else: return 'x'

    def submatch(self,decklist1,decklist2):
        deck1 = []
        deck2 = []
        for i in decklist1: deck1.append(i)
        for i in decklist2: deck2.append(i)
        rn.shuffle(deck1)
        rn.shuffle(deck2)
        points1 = 0
        points2 = 0
        for i in range(10):
            best_play1 = self.best_play(deck1[0:5])
            best_play2 = self.best_play(deck2[0:5])
            if best_play1[1] > best_play2[1]:
                points1 = points1 + 1
            elif best_play1[1] < best_play2[1]:
                points2 = points2 + 1
            for i in best_play1[0]: deck1.remove(i)
            for i in best_play2[0]: deck2.remove(i)
        if points1 > points2: return 0
        elif points1 < points2: return 1
        else: return 2
        
    def best_play(self,hand):
        combinations = list(itertools.combinations(hand,3))
        scores = []
        for i in range(len(combinations)):
            scores.append(self.score(combinations[i]))
        best_play = combinations[np.argmax(scores)]
        return (best_play,scores[np.argmax(scores)])
          
    def score(self,n):
        links1 = self.cards.links[n[0],n]
        score1 = self.cards.complete_set.power[n[0]]*(sum(links1))
        links2 = self.cards.links[n[1],n]
        score2 = self.cards.complete_set.power[n[1]]*(sum(links2))
        links3 = self.cards.links[n[2],n]
        score3 = self.cards.complete_set.power[n[2]]*(sum(links3))
        total_score = score1+score2+score3
        return total_score
    
class tournament:
    def __init__(self,game):
        self.game = game

    def make_tournament(self,agentset):
        n = len(agentset)
        x = np.zeros([n])
        self.ranking = pd.DataFrame(data = x, columns = ['points'])
        turns = np.ceil(math.log(n,2))
        for i in range(int(turns)):
            pairings = self.make_pairings()
            for j in range(len(pairings)):
                if pairings[j][1] == 'bye':
                    self.ranking.points[pairings[j][0]] += 3
                else:
                    result = self.game.match(agentset[pairings[j][0]],agentset[pairings[j][1]])
                    if result == 1:
                        self.ranking.points[pairings[j][0]] += 3
                    elif result == 2:
                        self.ranking.points[pairings[j][1]] += 3
                    else:
                        self.ranking.points[pairings[j][0]] += 1
                        self.ranking.points[pairings[j][1]] += 1
                
    def make_pairings(self):
        points = sorted(self.ranking.points.unique(),reverse = True)
        pairings = []
        leftover_player = []
        for i in range(len(points)):
            players = list(self.ranking.index[self.ranking.points == points[i]])
            if leftover_player:
                players.insert(0,leftover_player[0])
                leftover_player = []
            if len(players)%2 == 1:
                leftover_player.append(players[-1])
                players.pop(-1)
            while players:
                ran = rn.randint(1,len(players)-1)
                pairings.append((players[0],players[ran]))
                players.pop(ran)
                players.pop(0)
        if leftover_player:
            pairings.append((leftover_player[0],'bye'))
        return pairings

    def top_cards(self,agentset):
        top_cards = []
        n = len(self.ranking)
        number_of_decklist = int(np.ceil(n/10))
        top_players = list(self.ranking.sort_values('points',ascending = False)[0:number_of_decklist].index)
        for i in top_players:
            for j in agentset[i].decklist:
                top_cards.append(j)
        return top_cards


    


